import logging
import typing as t

from flask import session
from superset.security import SupersetSecurityManager

from . import bootstrap as cairn_bootstrap

logger = logging.getLogger(__name__)


OPENEDX_SSO_PROVIDER = "openedx"


class OpenEdxSsoSecurityManager(SupersetSecurityManager):
    def oauth_user_info(self, provider, response=None):
        """
        Identify the user
        """
        if provider == OPENEDX_SSO_PROVIDER:
            try:
                return self.get_user_info()
            except Exception as e:
                # Log exceptions, otherwise the stacktrace is swallowed by
                # flask_appbuilder.security.views.AuthOAuthView.oauth_authorized
                logger.exception(e)
                raise

    def get_user_info(self):
        """
        Make calls to the LMS API to fetch user information
        http://local.overhang.io:8000/api-docs/#/user/user_v1_me_read
        """
        username = self.get_lms_api("/api/user/v1/me")["username"]
        account = self.get_lms_api(f"/api/user/v1/accounts/{username}")

        # Fetch list of courses in which user is staff
        courses = [
            c["course_id"]
            for c in self.get_lms_api(
                f"/api/courses/v1/courses/?permissions=staff&username={username}"
            )["results"]
        ]
        if not courses:
            # User is not staff, entry is forbidden
            return {}

        # Create role, db, clickhouse db associated to user
        cairn_bootstrap.setup_user(username, course_ids=courses)

        # See flask_appbuilder.security.manager.BaseSecurityManager.auth_user_oauth for
        # valid keys
        return {
            "name": account["name"],
            "email": account["email"],
            "id": username,
            "username": username,
        }

    def get_lms_api(self, endpoint):
        """
        Make a call to the LMS API using the client app credentials.
        """
        return (
            self.appbuilder.sm.oauth_remotes[OPENEDX_SSO_PROVIDER].get(endpoint).json()
        )

    def _oauth_calculate_user_roles(self, userinfo) -> t.List[str]:
        """
        Override parent method to be able to create groups that match the user name.

        This is a bit hackish, but the cleanest solution we found.
        """
        roles = []
        for name in cairn_bootstrap.get_role_names(userinfo["username"]):
            role = self.find_role(name)
            if role:
                roles.append(role)
            else:
                logger.error("Could not find role: %s", name)

        # If user is already a member of one of these roles, preserve them.
        if user := self.find_user(username=userinfo["username"]):
            roles_to_preserve = ["Admin"]
            for role_to_preserve in roles_to_preserve:
                role = self.find_role(role_to_preserve)
                if role in user.roles:
                    roles.append(role)

        return roles

    def set_oauth_session(self, provider, oauth_response):
        """
        Store the oauth token in the session for later retrieval.
        """
        super().set_oauth_session(provider, oauth_response)

        if provider == OPENEDX_SSO_PROVIDER:
            session["oauth_token"] = oauth_response



import logging
import os
import typing as t

from cachelib.redis import RedisCache
from celery.schedules import crontab

from superset.extensions import security_manager
from superset.cairn import bootstrap as cairn_bootstrap
from superset.cairn import sso as cairn_sso

# https://superset.apache.org/docs/installation/configuring-superset
SECRET_KEY = "{{ CAIRN_SUPERSET_SECRET_KEY }}"
SQLALCHEMY_DATABASE_URI = "postgresql+psycopg2://{{ CAIRN_POSTGRESQL_USERNAME }}:{{ CAIRN_POSTGRESQL_PASSWORD }}@cairn-postgresql/{{ CAIRN_POSTGRESQL_DATABASE }}"

# Caddy is running behind a proxy: Superset needs to handle x-forwarded-* headers
# https://flask.palletsprojects.com/en/latest/deploying/proxy_fix/
# https://superset.apache.org/docs/installation/configuring-superset/#configuration-behind-a-load-balancer
ENABLE_PROXY_FIX = True

# Languages
# https://github.com/apache/superset/blob/dc575080d7e43d40b1734bb8f44fdc291cb95b11/superset/config.py#L324
available_languages = {
    "en": {"flag": "us", "name": "English"},
    "es": {"flag": "es", "name": "Spanish"},
    "it": {"flag": "it", "name": "Italian"},
    "fr": {"flag": "fr", "name": "French"},
    "zh": {"flag": "cn", "name": "Chinese"},
    "ja": {"flag": "jp", "name": "Japanese"},
    "de": {"flag": "de", "name": "German"},
    "pt": {"flag": "pt", "name": "Portuguese"},
    "pt_BR": {"flag": "br", "name": "Brazilian Portuguese"},
    "ru": {"flag": "ru", "name": "Russian"},
    "ko": {"flag": "kr", "name": "Korean"},
    "sl": {"flag": "si", "name": "Slovenian"},
}
{#- https://github.com/apache/superset/blob/master/docs/docs/contributing/translations.mdx#enabling-language-selection #}
enabled_language_codes = ["en"]
LANGUAGES = {}
if "{{ CAIRN_SUPERSET_LANGUAGE_CODE }}" in available_languages:
    enabled_language_codes.append("{{ CAIRN_SUPERSET_LANGUAGE_CODE }}")
    # Set the platform default language/locale
    BABEL_DEFAULT_LOCALE = "{{ CAIRN_SUPERSET_LANGUAGE_CODE }}"
for code in enabled_language_codes:
    LANGUAGES[code] = available_languages[code]

# Borrowed from superset/docker/pythonpath_dev/superset_config.py
REDIS_HOST = "{{ REDIS_HOST }}"
REDIS_PORT = "{{ REDIS_PORT }}"
# Be careful not to conflict with Open edX here
REDIS_CELERY_DB = {{ OPENEDX_CELERY_REDIS_DB + 2 }}
REDIS_CACHE_DB = {{ OPENEDX_CACHE_REDIS_DB + 2 }}

# Cache configuration
CACHE_CONFIG = {
    "CACHE_TYPE": "redis",
    "CACHE_DEFAULT_TIMEOUT": 60 * 60 * 24 * 1,  # 1 day default (in secs)
    "CACHE_KEY_PREFIX": "superset_data_cache",
    "CACHE_REDIS_URL": f"redis://{REDIS_HOST}:{REDIS_PORT}/{REDIS_CACHE_DB}",
}
DATA_CACHE_CONFIG = CACHE_CONFIG.copy()
FILTER_STATE_CACHE_CONFIG = CACHE_CONFIG.copy()
FILTER_STATE_CACHE_CONFIG.update({
    "CACHE_DEFAULT_TIMEOUT": 60 * 60 * 24 * 90,  # 90 days
    "REFRESH_TIMEOUT_ON_RETRIEVAL": True,
})
EXPLORE_FORM_DATA_CACHE_CONFIG  = CACHE_CONFIG.copy()
EXPLORE_FORM_DATA_CACHE_CONFIG.update({
    "CACHE_DEFAULT_TIMEOUT": 60 * 60 * 24 * 7,  # 7 days
    "REFRESH_TIMEOUT_ON_RETRIEVAL": True,
})
RESULTS_BACKEND = RedisCache(
    host=REDIS_HOST,
    port=REDIS_PORT,
    db=REDIS_CACHE_DB,
    key_prefix="superset_results",
)

{% if CAIRN_ENABLE_SSO %}
# Authentication
# https://superset.apache.org/docs/installation/configuring-superset/#custom-oauth2-configuration
# https://flask-appbuilder.readthedocs.io/en/latest/security.html#authentication-oauth
from flask_appbuilder.security.manager import AUTH_OAUTH
AUTH_TYPE = AUTH_OAUTH
OPENEDX_LMS_ROOT_URL = "{% if ENABLE_HTTPS %}https{% else %}http{% endif %}://{{ LMS_HOST }}"
OPENEDX_SSO_CLIENT_ID = "{{ CAIRN_SSO_CLIENT_ID }}"
if os.environ.get("FLASK_ENV") == "development":
    OPENEDX_LMS_ROOT_URL = "http://{{ LMS_HOST }}:8000"
    OPENEDX_SSO_CLIENT_ID = "{{ CAIRN_SSO_CLIENT_ID }}-dev"
OAUTH_PROVIDERS = [
    {
        "name": cairn_sso.OPENEDX_SSO_PROVIDER,
        "token_key": "access_token",
        "icon": "fa-right-to-bracket",
        "remote_app": {
            "client_id": OPENEDX_SSO_CLIENT_ID,
            "client_secret": "{{ CAIRN_SSO_CLIENT_SECRET }}",
            "client_kwargs": {"scope": "read"},
            "access_token_method": "POST",
            "api_base_url": f"{OPENEDX_LMS_ROOT_URL}",
            "access_token_url": f"{OPENEDX_LMS_ROOT_URL}/oauth2/access_token/",
            "authorize_url": f"{OPENEDX_LMS_ROOT_URL}/oauth2/authorize/",
        }
    }
]
CUSTOM_SECURITY_MANAGER = cairn_sso.OpenEdxSsoSecurityManager
# Update roles on login: this will cause all roles (except those that are preserved) to
# be ovewritten.
AUTH_ROLES_SYNC_AT_LOGIN = True
# Login will create user
AUTH_USER_REGISTRATION = True
{% endif %}

class CeleryConfig:  # pylint: disable=too-few-public-methods
    BROKER_URL = f"redis://{REDIS_HOST}:{REDIS_PORT}/{REDIS_CELERY_DB}"
    CELERY_IMPORTS = ("superset.sql_lab", "superset.tasks")
    CELERYD_LOG_LEVEL = "DEBUG"
    CELERYD_PREFETCH_MULTIPLIER = 1
    CELERY_ACKS_LATE = False
    CELERY_ANNOTATIONS = {
        "sql_lab.get_sql_results": {"rate_limit": "100/s"},
        "email_reports.send": {
            "rate_limit": "1/s",
            "time_limit": 120,
            "soft_time_limit": 150,
            "ignore_result": True,
        },
    }
    CELERYBEAT_SCHEDULE = {
        "email_reports.schedule_hourly": {
            "task": "email_reports.schedule_hourly",
            "schedule": crontab(minute=1, hour="*"),
        },
        "reports.scheduler": {
            "task": "reports.scheduler",
            "schedule": crontab(minute="*", hour="*"),
        },
        "reports.prune_log": {
            "task": "reports.prune_log",
            "schedule": crontab(minute=0, hour=0),
        },
    }


CELERY_CONFIG = CeleryConfig

# Avoid duplicate logging because of propagation to root logger
logging.getLogger("superset").propagate = False

# https://github.com/apache/superset/blob/master/RESOURCES/FEATURE_FLAGS.md
FEATURE_FLAGS = {
    # Enable dashboard embedding
    "EMBEDDED_SUPERSET": True
}

{{ patch("cairn-superset-settings") }}



from functools import lru_cache
import json
import logging
import os

import requests
from superset.extensions import db, security_manager
from superset.utils.database import get_or_create_db

logger = logging.getLogger(__name__)

# sql_lab is required in 2.1.0 for non-admin users to get access to sql queries
DEFAULT_ROLES = ["Gamma", "sql_lab"]

def setup_user(username: str, course_ids=None) -> None:
    """
    Create clickhouse DB, superset role, and superset DB associated to user.
    This role will have access to the database with the same name.
    """
    clickhouse_username = f"openedx-{username}"
    superset_db = f"openedx-{username}"
    superset_role = get_user_role_name(username)

    create_clickhouse_user(clickhouse_username)
    grant_clickhouse_row_based_access(clickhouse_username, course_ids=course_ids)
    create_superset_db(superset_db, clickhouse_username)
    create_superset_db_role(superset_role, superset_db)

def get_role_names(username: str) -> str:
    """
    Return all the role names normally associated to a user.
    """
    return DEFAULT_ROLES + [get_user_role_name(username)]

def get_user_role_name(username: str) -> str:
    """
    Return the user-specific role name associated to a user.
    """
    return f"openedx-{username}"

def create_superset_db(superset_database: str, clickhouse_username: str) -> None:
    """
    Create a database object with the right Clickhouse URI:

    - user: clickhouse_username
    - password: None
    - host/port: clickhouse host/port
    - database: database name

    User will be able to access the Clickhouse DB without any password, but should only
    be granted access to the right rows.
    """
    # https://superset.apache.org/docs/databases/clickhouse
    auth = get_clickhouse_credentials()
    uri = f"clickhousedb://{clickhouse_username}:@{auth['host']}:{auth['http_port']}/{auth['database']}"
    logger.info("Creating Superset DB: %s", uri)
    superset_db = get_or_create_db(superset_database, uri, always_create=True)
    db.session.add(superset_db)
    db.session.commit()


def create_superset_db_role(role_name: str, superset_database_name: str) -> None:
    """
    Create a role that has basic access permissions for a certain database.
    """

    def check_permission(permission_view) -> bool:
        """
        The list of all available permissions can be obtained from the admin role:

            print(security_manager.find_role("Admin").permissions)
        """
        permission_name = str(permission_view)
        if permission_name in [
            "can save on Datasource",
            "can sql json on Superset",
            "menu access on Datasets",
            # Modify "see table schema" dropdown in sql lab
            "can expanded on TableSchemaView",
            "can delete on TableSchemaView",
            "can post on TableSchemaView",
        ]:
            return True
        if permission_name.startswith(f"database access on [{superset_database_name}]"):
            return True
        if permission_name.startswith(f"schema access on [{superset_database_name}]"):
            return True
        return False

    # Create or update role with the same name as the user
    security_manager.set_role(role_name, check_permission)


def create_clickhouse_user(clickhouse_username):
    """
    Create a password-less clickhouse user with access to Clickhouse.
    """
    make_clickhouse_query(f"""CREATE USER IF NOT EXISTS '{clickhouse_username}';""")
    make_clickhouse_query(
        f"""GRANT CREATE TEMPORARY TABLE ON *.* TO '{clickhouse_username}';"""
    )


def grant_clickhouse_row_based_access(clickhouse_username, course_ids=None):
    """
    Grant row-based access to a Clickhouse user based on a selection of course IDs.

    When the list of course IDs is None, grant access to all courses.
    """
    if course_ids:
        condition = " OR ".join(
            [f"course_id = '{course_id}'" for course_id in course_ids]
        )
    else:
        condition = "1"
    # Find the list of tables to which the user should have access: all tables that do not start with "_"
    for table in make_clickhouse_query("SHOW TABLES").split("\n"):
        if not table.startswith("_"):
            make_clickhouse_query(
                f"""GRANT SELECT ON {table} TO '{clickhouse_username}';"""
            )
            make_clickhouse_query(
                f"""CREATE ROW POLICY OR REPLACE '{clickhouse_username}' ON {table} AS RESTRICTIVE FOR SELECT USING {condition} TO '{clickhouse_username}';"""
            )


def make_clickhouse_query(query):
    """
    Query Clickhouse by POSTing some content by http.
    """
    logger.info("Running Clickhouse query: %s", query)
    auth = get_clickhouse_credentials()
    clickhouse_uri = f"{auth['http_scheme']}://{auth['username']}:{auth['password']}@{auth['host']}:{auth['http_port']}/?database={auth['database']}"
    response = requests.post(clickhouse_uri, data=query.encode("utf8"), timeout=10)
    if response.status_code != 200:
        raise ValueError(
            f"An error occurred while attempting to post a query: {response.content.decode()}"
        )
    return response.content.decode("utf8").strip()


@lru_cache(maxsize=None)
def get_clickhouse_credentials():
    """
    Load the clickhouse credentials from file.
    """
    with open(
        os.path.join(os.path.dirname(__file__), "clickhouse-auth.json"),
        encoding="utf-8",
    ) as f:
        return json.load(f)
